/*
 *  toshbt.c - Toshiba Laptop Bluetooth Support
 *
 *
 *  Copyright (C) 2006 Mitch DSouza
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *  The devolpment page for this driver is located at
 *  http://0bits.com/toshbt
 *
 *  Credits:
 *      John Belmonte - based on toshiba_acpi kernel driver
 *
 *  TODO
 *
 */

#define MOD_VERSION	"1.0"
#define MOD_DESC "Toshiba Laptop Bluetooth Driver"
#define MOD_AUTHOR "Mitch DSouza"

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/proc_fs.h>
#include <asm/uaccess.h>

#include <acpi/acpi_drivers.h>

MODULE_AUTHOR(MOD_AUTHOR);
MODULE_DESCRIPTION(MOD_DESC);
MODULE_LICENSE("GPL");

#define MY_LOGPREFIX "toshbt: "
#define MY_ERR KERN_ERR MY_LOGPREFIX
#define MY_NOTICE KERN_NOTICE MY_LOGPREFIX
#define MY_INFO KERN_INFO MY_LOGPREFIX

/* Toshiba ACPI method paths */
#define METHOD_HCI_1		"\\_SB_.VALD.GHCI"
#define METHOD_HCI_2		"\\_SB_.VALZ.GHCI"

/* Toshiba HCI interface definitions
 *
 * HCI is Toshiba's "Hardware Control Interface" which is supposed to
 * be uniform across all their models.  Ideally we would just call
 * dedicated ACPI methods instead of using this primitive interface.
 * However the ACPI methods seem to be incomplete in some areas (for
 * example they allow setting, but not reading, the LCD brightness value),
 * so this is still useful.
 */

#define HCI_WORDS			6

/* operations */
#define HCI_SET				0xff00
#define HCI_GET				0xfe00

/* registers */
#define	HCI_SYSTEM_EVENT		0x0016

/* return codes */
#define HCI_SUCCESS			0x0000
#define HCI_FAILURE			0x1000
#define HCI_NOT_SUPPORTED		0x8000
#define HCI_EMPTY			0x8c00

/* the toshiba_acpi module */

/* acpi interface wrappers
 */

static int is_valid_acpi_path(const char *methodName)
{
	acpi_handle handle;
	acpi_status status;

	status = acpi_get_handle(NULL, (char *)methodName, &handle);
	return !ACPI_FAILURE(status);
}

static const char *method_hci /*= 0*/ ;

/* Perform a raw HCI call.  Here we don't care about input or output buffer
 * format.
 */
static acpi_status hci_raw(const u32 in[HCI_WORDS], u32 out[HCI_WORDS])
{
	struct acpi_object_list params;
	union acpi_object in_objs[HCI_WORDS];
	struct acpi_buffer results;
	union acpi_object out_objs[HCI_WORDS + 1];
	acpi_status status;
	int i;

	params.count = HCI_WORDS;
	params.pointer = in_objs;
	for (i = 0; i < HCI_WORDS; ++i) {
		in_objs[i].type = ACPI_TYPE_INTEGER;
		in_objs[i].integer.value = in[i];
	}

	results.length = sizeof(out_objs);
	results.pointer = out_objs;

	status = acpi_evaluate_object(NULL, (char *)method_hci, &params,
				      &results);
	if ((status == AE_OK) && (out_objs->package.count <= HCI_WORDS)) {
		for (i = 0; i < out_objs->package.count; ++i) {
			out[i] = out_objs->package.elements[i].integer.value;
		}
	}

	return status;
}

static acpi_status
hci_read2(u32 reg, u32* out1, u32* out2, u32* result)
{
        u32 in[HCI_WORDS] = { HCI_GET, reg, *out1, *out2, 0, 0 };
        u32 out[HCI_WORDS];
        acpi_status status = hci_raw(in, out);
        *out1 = out[2];
        *out2 = out[3];
        *result = (status == AE_OK) ? out[0] : HCI_FAILURE;
        return status;
}

static acpi_status
hci_write2(u32 reg, u32 in1, u32 in2, u32* result)
{
        u32 in[HCI_WORDS] = { HCI_SET, reg, in1, in2, 0, 0 };
        u32 out[HCI_WORDS];
        acpi_status status = hci_raw(in, out);
        *result = (status == AE_OK) ? out[0] : HCI_FAILURE;
        return status;
}

static char*
read_wk(char* p)
{
        u32 hci_result;
        u32 value, value2;
        value = 0;
        value2 = 0x0001;
        hci_read2(0x0056, &value, &value2, &hci_result);
        if (hci_result == HCI_SUCCESS) {
                p += sprintf(p,"wireless key:                  %d\n", value );
        } else {
                p += sprintf(p, "ERROR\n");
        }

        return p;
}

static char*
read_bt(char* p)
{
        u32 hci_result;
        u32 value, value2;
        value = 0;
        value2 = 0;
        hci_read2(0x0056, &value, &value2, &hci_result);
        if (hci_result == HCI_SUCCESS) {
                p += sprintf(p,"present:                  %d\n", (value & 0x0f)?1:0);
        } else {
                p += sprintf(p, "ERROR\n");
        }
        value = 0;
        value2 = 0x0001;
        hci_read2(0x0056, &value, &value2, &hci_result);
        if (hci_result == HCI_SUCCESS) {
                p += sprintf(p,"power:                    %d\n", (value & 0x80) >> 7);
                p += sprintf(p,"attach:                   %d\n", (value & 0x40) >> 6);
                p += sprintf(p,"wireless:                 %d\n", (value & 0x200) >> 9);
        } else {
                p += sprintf(p, "ERROR\n");
        }

        return p;
}

static unsigned long
write_bt(const char* buffer, unsigned long count)
{
        int value, value2;
        u32 hci_result;
        value2 = 0x0080;

        if (sscanf(buffer, " power : %i", &value) == 1 &&
                        value >= 0 && value <= 1) {
                hci_write2(0x0056, value, value2, &hci_result);
                if (hci_result != HCI_SUCCESS)
                        return -EFAULT;
        } else
        if (sscanf(buffer, " attach : %i", &value) == 1 &&
                        value >= 0 && value <= 1) {
                value2 = 0x0040;
                hci_write2(0x0056, value, value2, &hci_result);
                if (hci_result != HCI_SUCCESS)
                        return -EFAULT;
        } else
        if (sscanf(buffer, " wireless : %i", &value) == 1 &&
                        value >= 0 && value <= 1) {
                value2 = 0x0200;
                hci_write2(0x0056, value, value2, &hci_result);
                if (hci_result != HCI_SUCCESS)
                        return -EFAULT;
        } else {
                return -EINVAL;
        }

        return count;
}



/* common hci tasks (get or set one value)
 *
 * In addition to the ACPI status, the HCI system returns a result which
 * may be useful (such as "not supported").
 */

static acpi_status hci_write1(u32 reg, u32 in1, u32 * result)
{
	u32 in[HCI_WORDS] = { HCI_SET, reg, in1, 0, 0, 0 };
	u32 out[HCI_WORDS];
	acpi_status status = hci_raw(in, out);
	*result = (status == AE_OK) ? out[0] : HCI_FAILURE;
	return status;
}


typedef struct _ProcItem {
	const char *name;
	char *(*read_func) (char *);
	unsigned long (*write_func) (const char *, unsigned long);
} ProcItem;

/* proc file handlers
 */

static int
dispatch_read(char *page, char **start, off_t off, int count, int *eof,
	      ProcItem * item)
{
	char *p = page;
	int len;

	if (off == 0)
		p = item->read_func(p);

	/* ISSUE: I don't understand this code */
	len = (p - page);
	if (len <= off + count)
		*eof = 1;
	*start = page + off;
	len -= off;
	if (len > count)
		len = count;
	if (len < 0)
		len = 0;
	return len;
}

static int
dispatch_write(struct file *file, const char __user * buffer,
	       unsigned long count, ProcItem * item)
{
	int result;
	char *tmp_buffer;

	/* Arg buffer points to userspace memory, which can't be accessed
	 * directly.  Since we're making a copy, zero-terminate the
	 * destination so that sscanf can be used on it safely.
	 */
	tmp_buffer = kmalloc(count + 1, GFP_KERNEL);
	if (!tmp_buffer)
		return -ENOMEM;

	if (copy_from_user(tmp_buffer, buffer, count)) {
		result = -EFAULT;
	} else {
		tmp_buffer[count] = 0;
		result = item->write_func(tmp_buffer, count);
	}
	kfree(tmp_buffer);
	return result;
}

/* proc and module init
 */

static  struct proc_dir_entry*   toshiba_proc_dir;

static ProcItem proc_items[] = {
	{"wireless_key",	read_wk,	NULL},
	{"bluetooth",		read_bt,	write_bt},
	{NULL}
};

static acpi_status __init add_proc_entries(void)
{
	struct proc_dir_entry *proc;
	ProcItem *item;

	for (item = proc_items; item->name; ++item) {
		proc = create_proc_read_entry(item->name,
					      S_IFREG | S_IRUGO | S_IWUSR,
					      toshiba_proc_dir,
					      (read_proc_t *) dispatch_read,
					      item);
		if (proc)
			proc->owner = THIS_MODULE;
		if (proc && item->write_func)
			proc->write_proc = (write_proc_t *) dispatch_write;
	}

	return AE_OK;
}

static acpi_status __exit remove_proc_entries(void)
{
	ProcItem *item;

	for (item = proc_items; item->name; ++item)
		remove_proc_entry(item->name, toshiba_proc_dir);
	return AE_OK;
}

static int __init tosh_bt_init(void)
{
	acpi_status status = AE_OK;
	u32 hci_result;
	struct proc_dir_entry *de = acpi_root_dir;

	if (acpi_disabled)
		return -ENODEV;

	/* 
	 * Need to find the toshiba_proc_dir entry, but it's not an exported 
	 * symbol, so we search from the proc acpi_root_dir entry for it.
	 */
	for (de = de->subdir; de ; de = de->next) {
		if (de->namelen == 7 && !memcmp("toshiba", de->name, 7)) {
			toshiba_proc_dir = de;
			break;
		}
	}
	if (!toshiba_proc_dir) {
		printk("Can't find the proc dir for the toshiba_acpi module. "
		       "Maybe it's not loaded ?");
		return -ENODEV;
	}

	/* simple device detection: look for HCI method */
	if (is_valid_acpi_path(METHOD_HCI_1))
		method_hci = METHOD_HCI_1;
	else if (is_valid_acpi_path(METHOD_HCI_2))
		method_hci = METHOD_HCI_2;
	else
		return -ENODEV;

	/* enable event fifo */
	hci_write1(HCI_SYSTEM_EVENT, 1, &hci_result);

	/* 
	 * Poke the card up. I guess we wouldn't be loading this module if
         *  we didn't want the bluetooth to be up ?
	 */
	hci_write2(0x0056, 1, 0x0040, &hci_result);
	hci_write2(0x0056, 1, 0x0080, &hci_result);

	status = add_proc_entries();
	if (ACPI_FAILURE(status))
		remove_proc_entries();

	printk(MY_INFO MOD_DESC " %s\n",
	       MOD_VERSION);

	return (ACPI_SUCCESS(status)) ? 0 : -ENODEV;
}

static void __exit tosh_bt_exit(void)
{
	remove_proc_entries();
	return;
}

module_init(tosh_bt_init);
module_exit(tosh_bt_exit);
